                          (*       ENUM.PAS        *)


unit Enum;

interface

uses Enum_cons;

var
  outf, pfile: textfile;
  outfn, pname: shortstring;
    (* proname = project name, pname = printfilename *)

  last, undef, wrong, ready: Boolean;
  r0, r1: Sumsr;       (* remaining entries 0 and 1 in rows *)
  c0, c1: Sumsc;       (* remaining entries 0 and 1 in columns *)
  nul,n1x,nxx,rem1: Remember;   (* current number of  null,
                       1-x and x-x dyads, where x denotes undefined;
                       and number of remaining 1 entries *)
  impl00,impl10,impl11,impl01: Bomember; (* indicate whether for all 1x, 0x
                           and xx dyads, one entry implies the other:
                           e.g., impl10 means that every 1x dyad must become
                           a 10 dyad  *)
  choice1, choice2: Switch;
  def: Level;
    (* h is the number of current node;
       choice1[j] and choice2[j] give the coordinates of node number j;
       def[i,j] gives the node number where mat[i,j] is determined  *)
  halted,haltit: boolean;

procedure Initialise1;
procedure Initialise2;
procedure Findnext(var ready: Boolean);
procedure Foundleaf(good: Boolean; var last: Boolean);

implementation

uses SysUtils, Functen, Numbers, Zolib, Zoelib, ZOE_Form;

procedure Disnumbers;
var i, ErrInd, fgm: Byte;
    x: Real;
    r0,c0,rt1,ct1: Margin;
    totnumber: real;
    s: shortstring;
begin
  if gnrl then exit;
  for i:= 1 to f do r0[i] := rt[i];
  for i:= f+1 to fgmax do r0[i] := 0;
  for i:= 1 to g do c0[i] := ct[i];
  for i:= g+1 to fgmax do c0[i] := 0;
  if (f < g) then fgm:= g else fgm := f;
  ErrInd := 0;
  if full then begin
    try
      totnumber := bipart(r0,c0,fgm,ErrInd);
     except
      ErrInd := 5;
     end;
    if (ErrInd <= 0) then begin
      nicestring(totnumber,s);
      writeln(outf);
      writeln(outf,
        'The number of 0-1 matrices satisfying these constraints,');
      writeln(outf,'calculated by Sukhatme`s formula, is ',s,'.');
  (*  see Phil. Trans. Roy. Soc. London, A, vol. 237, 1937, pp. 386-389) *)
    end;
   end;
  if (not graph) then begin
    writeln(outf);
    x := Bender(f,g,rt,ct,strzer);
    writeln(outf,'Bender`s approximation for the total number of 0-1 matrices');
    writeln(outf,'with these marginal sums is ',exp(x):20:0);
    writeln(outf,'  (valid for large and sparse matrices...).');
(*   See Bender, Discrete Mathematics 10 (1974), 217-223 *)
    end;
  if (digr and (not mutfix) and calcnum) then begin
    for i:= f+1 to fgmax do rt1[i]:= 0;
    for i:= g+1 to fgmax do ct1[i]:= 0;
    for i:= 1 to f do rt1[i] := rt[i];
    for i:= 1 to g do ct1[i] := ct[i];
    try
      x := ndigraph(rt1,ct1,g,ErrInd);
     except
      ErrInd := 5;
      end;
    if (ErrInd <= 0) then begin
      writeln(outf,'The total number of digraphs for these degrees,');
      write(outf,'calculated by the Katz & Powell formula, is ');
      nicestring(x,s);
      writeln(outf,s,'.');
(*   See Katz and Powell, Proc. Am. Math. Soc. 5, 1954, 621-626
                        Ann. Math. Stat. 28, 1957, 442-448. *)
       end else begin
      writeln(outf);
      writeln(outf,
      'Request to calculate the total number of digraphs by Katz & Powell formula:');
      writeln(outf,
    'sorry - not possible because of insufficient memory (code ',ErrInd,').');
     end;
   end; (* if digr *)
end;  (* Disnumbers *)

function WrongMut: Boolean;
(* indicates that the number of mutuals cannot be equal to presmuts *)
begin
  if (mut[h] > presmuts) then WrongMut := True
    else begin
      if (rem1[h] <= n1x[h]) then WrongMut := (presmuts > (mut[h] + rem1[h]))
     else begin if (rem1[h] <= (n1x[h] + 2*nxx[h])) then
       WrongMut := (presmuts > (mut[h] + n1x[h] + ((rem1[h] - n1x[h]) div 2)))
      else WrongMut := (presmuts > (mut[h] + n1x[h] + nxx[h])); end; end;
end; (* WrongMut *)

procedure St0(i,j: Byte);
begin
  if (r0[h,i] > 0) then r0[h,i] := pred(r0[h,i]) else
    begin wrong := true; Exit; end;
  if (c0[h,j] > 0) then c0[h,j] := pred(c0[h,j]) else
    begin wrong := true; Exit; end;
  if ((def[j,i] <= h) and (i <> j) and (not mat[j,i])) then
                         nul[h] := succ(nul[h]);
  if (mutfix and (i<>j)) then begin
    if (def[j,i] <= h) then begin
      if mat[j,i] then n1x[h] := pred(n1x[h]); end
     else nxx[h] := pred(nxx[h]);
    if WrongMut then
      begin wrong := true; Exit; end;
   end; (* mutfix and i<>j *)
  def[i,j] := h;
  mat[i,j] := false;
end; (* St0 *)

procedure St1(i,j: Byte);
begin
  if (r1[h,i] > 0) then r1[h,i] := pred(r1[h,i]) else
    begin wrong := true; Exit; end;
  if (c1[h,j] > 0) then c1[h,j] := pred(c1[h,j]) else
    begin wrong := true; Exit; end;
  if ((def[j,i] <= h) and (i <> j) and mat[j,i]) then mut[h] := succ(mut[h]);
  if (mutfix and (i<>j)) then begin
    if (def[j,i] <= h) then begin
      if mat[j,i] then n1x[h] := pred(n1x[h]); end
     else begin
       nxx[h] := pred(nxx[h]);
       n1x[h] := succ(n1x[h]); end;
    if WrongMut then
      begin wrong := true; Exit; end;
   end; (* mutfix and i<>j *)
  rem1[h] := pred(rem1[h]);
  def[i,j] := h;
  mat[i,j] := true;
end; (* St1 *)

procedure Set0(i,j: Byte);
begin
(*  writeln(outf,'set0',i:2,j:2);*)
  St0(i,j);
  if wrong then Exit;
  if graph then begin
    if (def[j,i] >= hmax) then St0(j,i); if wrong then Exit; end else
   if mutfix then begin
    if (def[j,i] >= hmax) then begin
      if impl00[h] then St0(j,i) else
          begin if impl01[h] then St1(j,i); end; end;
    if wrong then Exit; end;
end; (* Set0 *)

procedure Set1(i,j: Byte);
begin
(*  writeln(outf,'set1',i:2,j:2); *)
  St1(i,j);
  if wrong then Exit;
  if graph then begin
    if (def[j,i] >= hmax) then St1(j,i); if wrong then Exit; end else
   if mutfix then begin
    if (def[j,i] >= hmax) then begin
      if impl10[h] then St0(j,i) else
          begin if impl11[h] then St1(j,i); end; end;
    if wrong then Exit; end;
end; (* Set1 *)

procedure CheckRem(i,j: Byte; var undef,wrong: Boolean);
(* This procedure checks if the remainder of row i and column j are defined
   and, if so, carries out the implied assignments *)
var rr1,rc1,rr0,rc0: Integer;
    jj: Byte;
begin
  rr1 := r1[h,i];        (* entries 1 for remainder row i *)
  rc1 := c1[h,j];        (* entries 1 for remainder column j *)
  rr0 := r0[h,i];        (* entries 0 for remainder row i *)
  rc0 := c0[h,j];        (* entries 0 for remainder column j *)
  undef := (rr1 > 0) and (rr0 > 0) and (rc1 > 0) and (rc0 > 0);
  (* undef = the remainder of row i and column j is not yet determined *)
  wrong := false;
  if not undef then begin
    if (rr1 <= 0) then begin
      for jj:= g downto 1 do
        begin if def[i,jj] >= hmax then Set0(i,jj); if wrong then Exit; end;
       end else if (rr0 <= 0) then begin
      for jj:= g downto 1 do
        begin if def[i,jj] >= hmax then Set1(i,jj); if wrong then Exit; end;
       end;
    if (rc1 <= 0) then begin
      for jj:= f downto 1 do
        begin if def[jj,j] >= hmax then Set0(jj,j); if wrong then Exit; end;
       end else if (rc0 <= 0) then begin
      for jj:= f downto 1 do
        begin if def[jj,j] >= hmax then Set1(jj,j);  if wrong then Exit; end;
       end;
    end;  (* not undef *)
end;  (* CheckRem *)

procedure Checkmut(var wrong: Boolean);
(* This procedure is called only in the MutFix mode,
   and carries out assignment that follow from the conditions
   where the further created number of mutuals is minimal
   or maximal, respectively *)
var i,j: Byte;
begin
  wrong := WrongMut;
  if wrong then Exit;
  if ((not impl10[h]) and (mut[h] >= presmuts)) then begin
  (* no more mutuals may be formed *)
     for i:= istar to g do
       for j :=g downto 1 do
         if ((def[i,j] >= hmax) and (def[j,i] <= h)) then
           begin if mat[j,i] then Set0(i,j); if Wrong then Exit; end;
     impl10[h] := true;
    end;
   if ((not impl00[h]) and (rem1[h] <= n1x[h]) and
      (presmuts >= (mut[h] + rem1[h]))) then begin
   (* all new entries 1 must be put in 1x dyads,
        so all 0x and xx dyads must become 00 dyads  *)
     for i:= istar to g do
       for j :=g downto 1 do
         if (def[i,j] >= hmax) then begin
          if (def[j,i] <= h) then
           begin if (not mat[j,i]) then Set0(i,j); if Wrong then Exit; end else
           if (def[j,i] >= hmax) then
             begin Set0(i,j); if (not wrong) then Set0(j,i);
               if Wrong then Exit; end;
          end;
     impl00[h] := true;
    end;
  if ((not (impl00[h] and impl11[h])) and (rem1[h] > n1x[h]) and
       (rem1[h] <= (n1x[h] + 2*nxx[h])) and
       (2*presmuts >= (2*mut[h] + n1x[h] + rem1[h]))) then begin
       (* all 1x dyads must become mutuals
           and all 0x dyads must become 00 dyads *)
    for i:= istar to g do
      for j :=g downto 1 do
        if ((def[i,j] >= hmax) and (def[j,i] <= h)) then
          begin if mat[j,i] then Set1(i,j) else Set0(i,j);
            if Wrong then Exit; end;
    impl00[h] := true; impl11[h] := true;
   end;
  if ((not impl11[h]) and (presmuts >= (mut[h] + n1x[h] + nxx[h]))) then
    begin
        (* all 1x and xx dyads must become mutuals *)
    for i:= istar to g do
      for j :=g downto 1 do
        if (def[i,j] >= hmax) then begin
          if (def[j,i] <= h) then
            begin if mat[j,i] then Set1(i,j); if Wrong then Exit; end else
            if (def[j,i] >= hmax) then
              begin Set1(i,j);
                if (not wrong) then Set1(j,i); if Wrong then Exit; end;
         end;
    impl11[h] := true;
   end;
end; (* CheckMut *)

procedure Initialise1;
var i,j,nf: Byte;
begin
  haltit := false;
  last := false;
  halted := false;
  for i:= 1 to fmax do begin
    for j:= 1 to gmax do begin
      mat[i,j] := false;
      strzer[i,j] := false;
     end;  (* j *)
    for j:= 0 to hmax do begin
      r0[j,i] := 0; r1[j,i] := 0;
     end; (* j *)
   end; (* i *)
 for i:= 1 to gmax do begin
   for j:= 0 to hmax do begin
     c0[j,i] := 0; c1[j,i] := 0;
    end; (* j *)
  end; (* i *)
  for i := 0 to hmax do begin
    mut[i] := 0;
    nul[i] := 0;
    n1x[i] := 0;
    impl00[i] := false;
    impl10[i] := false;
    impl11[i] := false;
    impl01[i] := false;
    for nf := 1 to numfmax do sumfn[i,nf] := 0;
   end; (* i *)
  for nf := 1 to numfmax do begin
    small[nf] := false;
    large[nf] := false;
    m4[nf] := 0;
    m3[nf] := m2;
    for j:= m1 to m2 do s[j,nf]:= 0.0;
    for j:= 1 to m01 do s0[j,nf] := 0.0;
   end; (* nf *)
  s00 := 0;
  Ndisconn := 0;
  h := 0;
  (* for use in EigenVector in Functen.pas : *)
  if (kmax > 15) then kmax := 15;
  printfileconn := false;
  f := 0; g := 0;
end;  (* Initialise1 *)

procedure Initialise2;
var i, j, ii: Integer;
procedure IfWrongStop;
begin
   if wrong then begin
     writeln(outf,'There is no 0-1 matrix with these marginals.');
     f := 0; g := 0;
     exit;
    end;
end; (* IfWrongStop *)
begin
  numf := Numfunc(fversion);
  nums := Numsumfn;
  if (numf > numfmax) then begin
    writeln(outf,'The number of functions evaluated is numf = ',numf,' .');
    writeln(outf,'This is greater than numfmax = ',numfmax,' .');
    writeln(outf,'Adapt numfmax. Program stops.');
    f := 0; g := 0;
    exit;
   end;
  if (nums > numfmax) then begin
    writeln
      (outf,'The number of functions primarily evaluated is nums = ',nums,' .');
    writeln(outf,'This is greater than numfmax = ',numfmax,' .');
    writeln(outf,'Adapt numfmax. Program stops.');
    f := 0; g := 0;
    exit;
   end;
  kmax := trunc(ln(2.147e9)/ln(g-1)); (* constant used in FUNCTEN *)
  if graph then begin
    impl00[0] := true; impl11[0] := true;
    impl10[0] := false; impl01[0] := false; end;
  for i:= 1 to fmax do begin
    for j:=1 to gmax do
      if strzer[i,j] then def[i,j] := 0 else def[i,j] := hmax; end;
  rem1[0] := sumc(rt);
  if mutfix then begin
    nxx[0] := (g*(g-1)) div 2;
   end;
  for i:= 1 to f do begin
    r0[0,i] := g-rzer[i]-rt[i];
    r1[0,i] := rt[i];
   end; (* i *)
  for i:= 1 to g do begin
    c0[0,i] := f-czer[i]-ct[i];
    c1[0,i] := ct[i];
   end; (* i *)
(*  if MutFix then
    begin istar := 1; jstar := 0; end ;
  else *)
  istar:=1; jstar:=g+1;
  (* Define those entries which are already determined by the given degrees *)
  Disnumbers;
  wrong := false;
  for ii := f+g downto 1 do begin
  (* This is in general a bit overdone; alternatively,
     it could be checked in this loop if subroutine CheckRem has led to
     any change and stop if it hasn't. But this would necessitate a
     more complicated subroutine CheckRem. In any case, f+g times can
     be proved to be enough. *)
    for i:= 1 to g do begin
      CheckRem(1,i,undef,wrong);
      IfWrongStop;
     end; (* i *)
    for i:= 1 to f do begin
      CheckRem(i,1,undef,wrong);
      IfWrongStop;
     end; (* i *)
    if (f = 0) then exit;
    if MutFix then CheckMut(wrong);
    IfWrongStop;
   end; (* ii *)
  if (f = 0) then exit;
  if printall then begin
    pname := proname+'.prn';
    pfname := pname;
    assign(pfile,pname);
    if FileExists(pname) then append(pfile) else rewrite(pfile);
    startpfile(pfile,rt,ct);
    printfileconn := true;
   end;

(* for checking :
   writeln(outf);
   writeln(outf,'Result of Initialise2');
   writeln(outf,'istar = ',istar,', jstar = ',jstar,', h = ',h,'.');
   for i:=1 to f do begin
     for j:=1 to g do
      if (def[i,j] < hmax) then
        if mat[i,j] then write(outf,' 1') else write(outf,' 0')
        else write(outf,' ?');
     writeln(outf);
    end;
   writeln(outf);
   for i:=1 to f do begin
     for j:=1 to g do write(outf,def[i,j]:2);
     writeln(outf);
    end;
   writeln(outf);  *)
end; (* Initialise2 *)

procedure Next(var i,j:Byte; var last:Boolean);
(* This procedure sets (i,j) to its succeeding value *)
var h: Byte;
begin
{  case MutFix of
  false:}  begin j := pred(j);  (* i.e. j-1 *)
                if (j <= 0) then begin
                  i := succ(i); j := g; end;
                last := (i > f);
          end;
{  true:
  begin if (j > i) then
                 begin h := i; i := j; j := h; last := false; end
               else
                begin if (j >= pred(i)) then
                  begin j := g; last := (i >= f); end
                 else
                  begin h := i; i := j; j := pred(h); last := false; end;
              end; end;
   end; (* case *) }
end; (* Next *)

(* function Check: Boolean;
var temp: Boolean;
  j : Byte;
begin
  temp := true;
  for j := f downto 1 do
    temp := temp and (r1[h,j] = 0) and (r0[h,j] = 0);
  for j := g downto 1 do
    temp := temp and (c1[h,j] = 0) and (c0[h,j] = 0);
  Check := temp;
  if not temp then begin gotoxy(60,24); write('Wrong Check'); end;
end ;  *)

procedure Foundleaf(good: Boolean; var last: Boolean);
label Earlier;
var i,j,nf: Byte;
    llast, GoOn: Boolean;
    rff: rFuncvalue;
begin
Earlier:
  if good then begin
    if connect then begin
    (* Check whether the graph constructed is connected *)
      GoOn := connected(mat);
      if not GoOn then Inc(Ndisconn);
     end  else GoOn := true;
    if GoOn then
    begin Update(rff);
      if ((s00 <= 100) or ((s00 <= 900) and (s00 mod 100 = 0))
        or (s00 mod 1000 = 0) )then
        ZOForm.shownumber(s00);
      if (s00 = 2147483647) then begin
        writeln(outf,' Before finishing the enumeration, the program has reached');
        writeln(outf, 'the number of 2,147,483,647 matrices, which is the maximum integer in Delphi.');
        writeln(outf,'Therefore the program must now stop.');
        haltit := true;
        exit;
       end;
      if printall then Listmat(outf,pfile,rff);
      if (rem1[h] > 0) then begin
(* just a check; this should never happen *)
        writeln(outf); writeln(outf,'Error: Not all entries one have a place !');
       end;
      end; (* if GoOn *)
     end; (* if good *)
  last := (h <= 0);
  if last then Exit;
  wrong := false;
  istar := choice1[h];
  jstar := choice2[h];
  h := pred(h);
  Set1(istar,jstar);
  if wrong then begin good := false; goto Earlier; end;
  i := istar; j := jstar;
  Next(i,j,llast);
  while not llast do begin
      if (def[i,j] > h) then def[i,j] := hmax; Next(i,j,llast); end;
  CheckRem(istar,jstar,undef,wrong);
  if ((not wrong) and MutFix) then CheckMut(wrong);
  if wrong then
    begin good := false; goto Earlier; end;
  Contr(h);
{  for nf := nums downto 1 do
    sumfn[h,nf] := sumfn[h,nf] + Contr(nf); }
 (* This is an updating assignment; see FUNCTEN.PAS *)
  end; (* Foundleaf *)

procedure Findnext(var ready: Boolean);
var i,j,hh : Byte;
    indef : Boolean;
begin
  hh := h;
  repeat
    Next(istar,jstar,ready);
    (* (ready = true) means that the total adjacency matrix is determined *)
    if ready then  Exit;
    if wrong then Exit;
    if (def[istar,jstar] >= hmax) then CheckRem(istar,jstar,indef,wrong)
       else indef := false;
    if (MutFix and (not wrong) and (def[istar,jstar] >= hmax))
      then CheckMut(wrong);
    indef := (indef and (def[istar,jstar] >= hmax));
    if wrong then Exit;
    if (not indef) then
{      for i := nums downto 1 do
        sumfn[h,i] := sumfn[h,i] + Contr(i);}
     Contr(h);
   until indef;
  if (succ(h) >= hmax) then
    begin writeln(outf); write(outf,'More than ');
          writeln(outf,pred(hmax),' levels in the enumeration tree!');
          writeln(outf,'Too many for this program.');
          Haltit := true;
          exit;
         end
     else h:= succ(h);
  mut[h] := mut[hh];
  nul[h] := nul[hh];
  rem1[h] := rem1[hh];
  if mutfix then begin
    n1x[h] := n1x[hh];
    nxx[h] := nxx[hh];
    impl00[h] := impl00[hh];
    impl10[h] := impl10[hh];
    impl11[h] := impl11[hh];
    impl01[h] := impl01[hh];
   end;
  for i:= f downto 1 do begin
    r1[h,i] := r1[hh,i];
    r0[h,i] := r0[hh,i];
   end;
  for i:= g downto 1 do begin
    c1[h,i] := c1[hh,i];
    c0[h,i] := c0[hh,i];
   end;
  for i := nums downto 1 do sumfn[h,i] := sumfn[hh,i];
  Set0(istar,jstar);        (* indef was false, so this cannot lead to wrong *)
(*  for i := nums downto 1 do sumfn[h,i] := sumfn[h,i] + Contr(i); *)
  Contr(h);
  choice1[h] := istar;
  choice2[h] := jstar;

(* for checking :
   writeln(outf);
   writeln(outf,'Result of Findnext');
   writeln(outf,'istar = ',istar,', jstar = ',jstar,', h = ',h,'.');
   for i:=1 to f do begin
     for j:=1 to g do
       if (def[i,j] < hmax) then
         if mat[i,j] then write(outf,' 1') else write(outf,' 0')
        else write(outf,' ?');
     writeln(outf);
    end;
   writeln(outf);
   for i:=1 to f do begin
     for j:=1 to g do write(outf,def[i,j]:2);
     writeln(outf);
    end;
   writeln(outf);  *)
end;  (* Findnext *)

begin    (* unit Enum *)
end.
