                 (* SIM.PAS  *)

unit Sim;

interface

Uses
  Simul, ZOMat, Triads;

(* const fmax, gmax, m1, m2, numfmax, numlmax, numl2max defined in Simul *)

type
(*  Adjac = Object array[1..fmax,1..gmax] of Boolean;    adjacency matrix
                                          defined in unit ZOMat *)
(* Defined in Simul:
  Marginc = array[1..gmax] of Byte;
  Marginr = array[1..fmax] of Byte;
  Res  = array[1..numfmax,m1..m2] of Dreal;
  ResB = array[1..numfmax] of Boolean;
  ResI = array[1..numfmax] of Integer;
  ResL = array[1..numl2max] of Dreal;
  ResL2 = array[1..numlmax,1..numlmax] of Dreal;
  Funcvalue1 = array[1..numfmax] of Integer;
  Funcvalue2 = array[1..numl2max] of Real; *)

  Res1 = array[m1..m2] of Real; (* used in Giveresult etc. *)
  Ress = array[1..70] of Real;   (* used in Barchart *)

var
  f, g, numl, mode: Byte;
  nrun, nsuc, ntry, plusrun: Longint;
  rt: Marginr; ct: Marginc; (* prescribed row margins *)
  mutzero,presmuts: Integer;
         (* origin for, and prescribed numbers of mutuals *)
  strzer: Adjac; (* prescribed matrix of structural zero indicators *)
  s, s2, c: Res;             (* sums and sums of squares of results *)
  sl, s2l, cl: ResL;         (* idem *)
  ssl: ResL2;                (* sums of results *)
  sp, sp2: Dreal;            (* sum and sum of squares of npra *)
  meanp,varp: Dreal;         (* final results *)
  mean,sd,cummean,cumsd: Res;(* final results *)
  meanl,sdl: ResL;           (* final results *)
  covl: ResL2;               (* final results *)
  large, small: ResB;        (* final results *)
  full, digr, graph, gnrl, mutfix, connect, rorder,
    MolloyReed, Readinfile: Boolean;
                             (* mode of adjacency matrix *)
  weit: TWeights;            (* weights for lin comb of triad counts *)
  outname, pname: string;    (* filename for output *)
  infile, dumpfile: Text;    (* file for input etc. *)
  errtext, warntext: string;

implementation;


procedure Giveresult;
var rang,barnum,ii,j: Integer;
    h, i: Byte;
    m3,m4: ResI;
    mean1, sd1: Res1;
    summean,zero: Ress;
    sdp: Real;
    chaar: Char;
    hzero: integer;
begin

(* First results for the probability distribution of the numf functions *)
  Extremes(mean,m3,m4);
  sdp := root(sp2);
  for h:=1 to numf do begin
    clrscr;
    if (h=1) then hzero := mutzero else hzero := 0;
    writeln(outf,Tex(h,0,fversion,graph,digr));
    write(outf,'Estimated mean               ', hzero + mean[h,-1]:8:4);
    writeln(outf,'     (standard error ', sd[h,-1]:8:4, ' )');
    writeln(outf,'Estimated variance           ', mean[h,-2]:8:4);
    writeln(outf,'Estimated standard deviation ', root(mean[h,-2]):8:4);
    if small[h] then begin
      writeln(outf,'A function value was smaller than ',hzero,'.');
      writeln(outf,'All smaller values were set to ',hzero,'.');
     end; (* if *)
    if large[h] then begin
      writeln(outf,'A function value was larger than ',m2+hzero,' .');
     end; (* if *)
    rang := ((m4[h]-m3[h]) div 70) + 1;
    if (rang <= 0) then rang := 1;
    barnum := (m4[h]-m3[h]+1) div rang;
    if rang*barnum < m4[h]-m3[h]+1 then barnum := barnum + 1;
    j := m3[h]-1;
    for ii:=1 to barnum do begin
      summean[ii] := 0.0;
      zero[ii] := 0.0;
      for i := 1 to rang do begin
        j := j+1;
        if (j <= m2) then summean[ii] := summean[ii]+mean[h,j];
       end; (* i *)
      end;(* ii *)
    Barchart(outf,1,rang,barnum,m3[h]+hzero,summean,zero);
    for ii :=0 to m2 do
      begin mean1[ii] := cummean[h,ii]; sd1[ii] := cumsd[h,ii]; end;
    writeln(outf,'Minimum value found ',m3[h],'; maximum ',m4[h],' .');
    writeln(outf);
    writeln
     (outf,'Estimated cumulative ( <= ) probabilities with standard errors.');
    Report1(outf,m3[h],m4[h],mean1,sd1,hzero);
    for ii :=0 to m2 do
      begin mean1[ii] := mean[h,ii]; sd1[ii] := sd[h,ii]; end;
    writeln(output,'Estimated probabilities with standard errors.');
    Report1(output,m3[h],m4[h],mean1,sd1,hzero);
   end; (* h *)
  Report2(outf,meanl,sdl,covl);
end;      (* Giveresult *)


procedure Result0(var ff: Text);
var i: Byte;
begin
  writeln(ff,'Number of runs ', nrun, ';  number of trials ', ntry, ' .');
  writeln(ff,'Row totals : ');
  for i:=1 to f do begin
    write(ff,rt[i]:3);
    if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
  writeln(ff);
  writeln(ff,'Column totals : ');
  for i:=1 to g do begin
    write(ff,ct[i]:3);
    if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
  writeln(ff); writeln(ff);
  if full then writeln(ff,'No structural zeros.');
  if digr then writeln(ff,'Directed graph.');
  if graph then writeln(ff,'Undirected graph.');
  if connect then begin
    write(ff,'Restriction to ');
    if digr then write(ff,'weakly ');
    write(ff,'connected ');
    if digr then write(ff,'di');
    writeln(ff,'graphs.');
   end; (* if *)
  if mutfix then writeln(ff,'Number of mutuals fixed.');
  if MolloyReed then writeln(ff,'Molloy-Reed algorithm was used.');
  if not UseResfile then begin
   if (root(varp*nrun) > 2.0) then begin
    write(ff,'The standard deviation of 1/probabilities is ');
    writeln(ff,root(varp*nrun):10:2);
    writeln(ff,'If this standard deviation is very high, then the simulation procedure');
    writeln(ff,'can be unreliable.');
    writeln(ff,'(50 is high; 100 is very high.)');
    writeln(ff);
    end;
   end;
end; (* Result0 *)

begin    (* Main Program *)
  Simulate(true,f,g,mode,strzer,presmuts,rt,ct,weit,numl,nrun,
           errtext,warntext,mutzero,nsuc,ntry,
           small,large,s,s2,c,sp,sp2,sl,s2l,cl,ssl);
  if (errtext > '') then NoMore(errtext,printall,useResfile,outf,pfile,resfile);

  repeat
    if (not UseResfile) then begin
      Calcresult(nrun, numf, numl, numl2,
                 sp, sp2, s, s2, c, sl, s2l, cl, ssl,
          meanp,varp, mean,sd,cummean,cumsd, meanl,sdl,covl);
      Result0(outf);
      GiveResult;
     end else begin
      Result0(outf);
     end;
    if UseResfile then begin
      writeln(outf,'Intermediate results written to file SIM.PQR (drive as given in SIM.DD).');
      writeln(outf,'For obtaining processed results, use program SIMAFTER or SIMPLUS.');
     end; (* if *)
    write('How more runs do you wish (0 = stop) ?');
    readln(plusrun);
    nrun := nrun + plusrun;
    if (plusrun >0) then
    Simulate(false,f,g,mode,strzer,presmuts,rt,ct,weit,numl,nrun,
              errtext,warntext,mutzero,nsuc,ntry,
              small,large,s,s2,c,sp,sp2,sl,s2l,cl,ssl);
    if (errtext > '') then NoMore(errtext,printall,useResfile,outf,pfile,resfile);
   until plusrun = 0 ;
  close(outf);
  if printall then close(pfile);
end.
