                            (* SIMPLUS.PAS *)

(* This unit calculates results for the program ZO_SIM for the case
   that intermediate results have been written to the file SIM.PQR,
   rather than stored in memory.
   This option is chosen in ZO_SIM if NumFunc is greater than NumfMax. *)


unit SimPlus;

interface

procedure SimPlusCalculations;

implementation

uses SysUtils, ZOS_Form, ZOLib, Simul;

const
  numfmax = 200;  (* Maximum number of functions that can be evaluated *)

type
(* for calculating function values : *)
  Funcvalue1 = array[1..numfmax] of Integer;
(* for bookkeeping of results: *)
  Res  = array[m1..m2] of real;


var
(* for input/output : *)
  nsuc: Longint;
  numf: integer;                (* number of evaluated statistics *)
  succes: Boolean;
  s, s2, c: Res;             (* sums and sums of squares of results *)
  sp, sp2: real;            (* sum and sum of squares of npra *)
  meanp,varp: real;         (* final results *)
  mean,sd,cummean,cumsd: Res1;(* final results *)
  large, small: Boolean;        (* final results *)
  Tx: array[1..numfmax] of shortstring;



procedure Init(var numf: integer);
var i: integer;
begin
  readln(resfile,numf);
  for i:=1 to numf do readln(resfile,Tx[i]);
end; (* Init *)

procedure Initi(var nsuc: longint;
     var s, s2, c: Res; var sp, sp2: real);
(* Initialisation for the bookkeeping procedures *)
var ii: integer; tt: shortstring;
begin
  nsuc := 0;
    small := false; large := false;
    for ii:= m1 to m2 do
      begin
        s[ii]:= 0.0;
        s2[ii] := 0.0;
        c[ii] := 0.0;
      end; (* ii *)
  sp:= 0.0;
  sp2 := 0.0;
  readln(resfile,tt);
  while not (tt = '@') do begin
    readln(resfile,tt);
   end; (* while *)
  (* now resfile is positioned where the data start *)
end;

procedure Update(var succes: boolean; h: integer);
(* Called in Simulate after each succesful matrix construction *)
var ff: Integer;
    hh: integer; fv: Funcvalue1;
    x, xx, y, yy: real;
begin
  succes := false;
  if (not eof(resfile)) then read(resfile,x) else exit;
  xx := sqr(x);
  sp  := sp + x;
  sp2 := sp2 + xx;
  for hh := 1 to numf do
    if (not eof(resfile)) then read(resfile,fv[hh]) else exit;
  if (not eof(resfile)) then readln(resfile);
  succes := true;
    ff := fv[h];
    if (ff < 0) then begin
      ff := 0; small := true;
     end; (* if *)
    if (ff <= m2) then begin
      s[ff]  := s[ff] + x;
      s2[ff] := s2[ff] + xx;
      c[ff]  := c[ff] + xx;
     end else large := true;
    y  := ff*x;
    yy := ff*y;
    s[-1]  := s[-1] + y;
    s[-2]  := s[-2] + yy;
    s2[-1] := s2[-1] + sqr(y);
    s2[-2] := s2[-2] + sqr(yy);
    c[-1]  := c[-1] + y*x;
    c[-2]  := c[-2] + yy*x;
end;     (* Update *)

procedure Calcresult(nrun: Longint;
(* input arrays and dreals : *)
        var sp,sp2: real; var s,s2,c: Res;
(* output arrays and dreals : *)
        var meanp,varp: real; var mean,sd,cummean,cumsd: Res1);
var ii: Integer;
    covi, cums1, cums2, cumc: Real;
begin
(* Correspondence with formulae in Snijders, Psychometrika 1991, are given *)
  meanp := sp/nrun;            (* (1/n) Sum(r(X_i)) *)
  varp := (sp2/nrun - sqr(meanp))/(nrun-1);  (* (sigma^2_r) / n *)
    cums1 := 0.0;
    cums2 := 0.0;
    cumc  := 0.0;
    for ii:= m1 to m2 do
      begin
(* for phi = indicator function of Funct[h] = ii : *)
        mean[ii] := s[ii]/nrun;          (* (1/n) Sum(phi(X_i)r(X_i)) *)
        covi  := (c[ii]/nrun  - meanp*mean[ii])/(nrun-1);
                            (* rho * sigma^2_r * sigma^2_phi,r / n *)
        if (ii >= 0) then begin
          cums1    := cums1 + mean[ii];
          cums2    := cums2 + s2[ii]/nrun;
          cumsd[ii] := (cums2 - sqr(cums1))/(nrun-1);
          cumc     := cumc + covi;
          cummean[ii] := cums1/meanp;
          cumsd[ii]   :=
 root(cumsd[ii] - 2*cumc*cummean[ii] + varp*sqr(cummean[ii]))/meanp;
         end; (* if ii *)
        sd[ii] := (s2[ii]/nrun - sqr(mean[ii]))/(nrun-1);
                                          (* sigma^2_phi,r / n *)
        mean[ii] := mean[ii]/meanp;   (* mu-hat *)
        sd[ii] :=
   root(sd[ii] - 2*covi*mean[ii] + varp*sqr(mean[ii]))/meanp ;
      end; (* ii *)
    mean[-2] := mean[-2] - sqr(mean[-1]);
end;      (* Calcresult *)

procedure Extremes(var mean: Res1; var m3,m4: integer);
var i: Integer;
begin
    m3 := m2; m4 := 0;
    for i:=0 to m2 do
      if (mean[i] > 1e-30) then begin
        if (m3 > i) then m3 := i;
        if (m4 < i) then m4 := i;
       end; (* if *)
end; (* Extremes *)

procedure Report1(var fl: Textfile; m3,m4:Integer; var mean,se: Res1);
(* mean and se are estimates and their standard errors *)
var i,row,nrows,rowmin,rowleng: Integer;
begin
   if (m3 > m4) then begin
     writeln(fl,'Minimum index is ',m3,' while maximum is ',m4,' .');
     writeln(fl,'No report.'); writeln(fl);
     Exit;
    end; (* if *)
   nrows := ((m4-m3) div 10) + 1;
   for row := 1 to nrows do begin
     rowmin := m3 + (row-1)*10 - 1;
     rowleng := m4 - m3 - row*10 + 11;
     if (rowleng > 10) then rowleng := 10;
     write(fl,'  ',rowmin+1:3);
     for i:= 2 to rowleng do write(fl,rowmin + i:7); writeln(fl);
     for i:= 1 to rowleng do write(fl,mean[rowmin+i]:7:4); writeln(fl);
     for i:= 1 to rowleng do write(fl,se[rowmin+i]:7:4); writeln(fl);
     writeln(fl);
    end; (* rows *)
  writeln(fl);
end (* Report1 *);

procedure Giveresult(h: integer);
var rang,barnum,i,ii,j: Integer;
    m3,m4: integer;
    mean1, sd1: Res1;
    summean,zero: Ress;
begin
(* First results for the probability distribution of the numf functions *)
  Extremes(mean,m3,m4);
    if ((h <= 1) and (root(varp*nsuc) > 2.0)) then begin
      write(outf,'The standard deviation of 1/probabilities is ');
      writeln(outf,root(varp*nsuc):10:2);
      writeln(outf,'If this standard deviation is very high, then the simulation procedure');
      writeln(outf,'can be unreliable.');
      writeln(outf,'(50 is high; 100 is very high.)');
      writeln(outf);
     end;
    heading(3,outf,tx[h]);
    write(outf,'Estimated mean               ', mean[-1]:8:4);
    writeln(outf,'     (standard error ', sd[-1]:8:4, ' )');
    writeln(outf,'Estimated variance           ', mean[-2]:8:4);
    writeln(outf,'Estimated standard deviation ', root(mean[-2]):8:4);
    if small then begin
      writeln(outf,'At least one function value was smaller than 0.');
      writeln(outf,'All negative values were set to 0.');
     end; (* if *)

    if (m3 < m4) then begin
      if large then begin
        writeln(outf,'At least one function value was larger than ',m2,' .');
       end; (* if *)
      rang := ((m4-m3) div 70) + 1;
      if (rang <= 0) then rang := 1;
      barnum := (m4-m3+1) div rang;
      if rang*barnum < m4-m3+1 then barnum := barnum + 1;
      j := m3-1;
      for ii:=1 to barnum do begin
        summean[ii] := 0.0;
        zero[ii] := 0.0;
        for i := 1 to rang do begin
          j := j+1;
          if (j <= m2) then summean[ii] := summean[ii]+mean[j];
         end; (* i *)
        end;(* ii *)
      Barchart(outf,1,rang,barnum,m3,summean,zero);
      writeln(outf);
      for ii :=0 to m2 do
        begin mean1[ii] := cummean[ii]; sd1[ii] := cumsd[ii]; end;
      writeln(outf,'Minimum value found ',m3,'; maximum ',m4,' .');
      writeln(outf);
      writeln
       (outf,'Estimated cumulative ( <= ) probabilities with standard errors.');
      Report1(outf,m3,m4,mean1,sd1);
      for ii :=0 to m2 do
        begin mean1[ii] := mean[ii]; sd1[ii] := sd[ii]; end;
      writeln(outf);
      writeln(outf,'Estimated probabilities with standard errors.');
      Report1(outf,m3,m4,mean1,sd1);
     end; (* if m3 < m4 *) 
end;      (* Giveresult *)

procedure SimPlusCalculations;
var i: integer;
  fna: shortstring;   (* filename *)
begin
  fna := proname+'.pqr';
  if not FileExists(fna) then begin
    writeln(outf,'File ',fna,' cannot be found.');
    writeln(outf,'The reason may be that a wrong directory is searched.');
    writeln(outf,'Take care that temporary files are in the current directory.');
    writeln(outf,'(In STOCNET, look in Options - Directories.)');
   end else begin
    assign(resfile,fna);
    reset(resfile);
    writeln(outf,'Results from file ',fna,' are processed.');
    Init(numf);
    for i:=1 to numf do begin
      reset(resfile);
      Initi(nsuc, s, s2, c, sp, sp2);
      while not (eof(resfile)) do begin
        Update(succes,i);
        if succes then Inc(nsuc);
       end;
      close(resfile);
      Calcresult(nsuc, sp,sp2, s,s2,c,meanp,varp, mean,sd,cummean,cumsd);
      Giveresult(i);
     end; (* for i *)
   end; (* if FileExists *)
end; (* simpluscalculations *)

begin (* unit SimPlus *)
end.