                       (* ZOSIMLIB.PAS *)

unit Zosimlib;

interface

uses simul;

procedure readinfile(var outf: textfile; proname: shortstring);
procedure Giveresult(var outf: textfile);
procedure Result0(var ff: Text; mode: byte);

implementation

Uses SysUtils, ZOMat, ZOLib, Triads, FunctSim;

function parsestring(rs: shortstring; var x: real): boolean;
var i: integer; rs1: shortstring;
 ci: char;
begin
 rs1 := '';
 for i:=1 to length(rs) do begin
   ci := rs[i];
   if (ci in ['+','-','.','1','2','3','4','5','6','7','8','9','0'])
     then rs1 := rs1 + ci;
  end;
 if (rs1 = '') then begin
  x := 0.0;
  parsestring := false;
 end else begin
   try
     x := strtofloat(rs1);
     parsestring := true;
    except
     x := 0.0;
     parsestring := false;
    end; (* try *)
  end; (* else *)
end; (* parsestring *)

procedure readinfile(var outf: textfile; proname: shortstring);
(* First part is almost identical to readinfile in ZOElib.
   changes: inclusion of MolloyReed; no calcnum; strzer handled differently;
   numversion has arguments.
   If an error is encountered, this procedure is exited with f = g = 0 . *)
var infn, datfn, ert, ts: shortstring;
 inf, datf: textfile;
 dum, dumij, i, j, fgm, option, iend, jend, CheckNumber: integer;
 x: real;
 good, datmatrix, symm, anycrit: boolean;
 crit: array[1..numlmax] of boolean;
 rzer: marginr; czer: marginc;
 full, digr, graph, gnrl, mutfix, conn, rorder,
    MolloyReed, SameDegrees: Boolean;   (* mode of adjacency matrix *)
 dmat: adjac;
 f1: Funcvalue1;
 f2: Funcvalue2;
 rf1: rFuncvalue1;
 tc: TriadCountType;
begin

(*
dimension f
dimension g
graph mode: 1 = full, 2 = digr, 3 = graph, 4 = arbitrary;
      +10 for Molloy-Reed algorithm
write mode: 0 = no print, 1 = print ("printall")
0 for marginal totals file, 1 for adjacency file
file name
connected mode: 0 = no, 1 = yes
if digr then mutuals mode: 0 = no, 1 = yes
if mutuals then presmuts
if gnrl then file name with structural zeros     <this line may be absent>
version for evaluated statistics: truncated if not between 1 and NumVersion
nrun = number of simulation
numl = number of linear combinations of triad counts (between 0 and numlmax = 18)
option 1 or 2
  if option = 2 and numl >= 1: file name for weights of linear combination
  this must be a file with numl lines and in each line 17 real numbers,
  separated by blanks.
*)
  printfileconn := false;
  heading(2,outf,'Reading input files.');
  infn := proname+'.in';
  if (not FileExists(infn)) then begin
    errtext := 'File '+infn+' is not found.';
    f := 0; g := 0;
    exit;
   end;
  assign(inf,infn);
  reset(inf);
  writeln(outf,'input file ',infn);
  readln(inf,f);
  readln(inf,g);
  writeln(outf,f,' rows, ',g,' columns.');
  if ((f <= 1) or (g <= 1)) then begin
    writeln(outf,'This is not meaningful.');
    errtext := 'Numbers of rows and columns not meaningful.';
    f := 0; g := 0;
   end;
  readln(inf,dum);
  full := false;
  digr := false;
  graph := false;
  gnrl := false;
  MolloyReed := (dum > 10);
  if MolloyReed then dum := dum-10;
  case dum of
    1: full := true;
    2: digr := true;
    3: graph := true;
    4: gnrl := true;
    else begin
      if MolloyReed then dum := dum+10;
      writeln(outf,'Error: code ',dum,' not allowed as type of matrix.');
      Errtext := 'Error in .in file.';
      f := 0; g := 0;
     end;
   end; (* case *)
  if (graph or digr) then begin
    g := f;
    if (f <= 2) then begin
      writeln(outf,'ZO analysis for a (di)graph with ',f,' vertices is not meaningful.');
      f := 0; g := 0;
     end;
   end;
  readln(inf,dum);
  printall := (dum > 0);
  readln(inf,dum);
  readln(inf,datfn);
  if (not FileExists(datfn)) then begin
    errtext := 'Data file '+datfn+' is not found.';
    f:=0; g := 0;
   end;
  if (f*g = 0) then exit;

(* read data file: *)
  assign(datf,datfn);
  reset(datf);
  if (dum <= 0) then begin (* marginal totals file *)
    datmatrix := false;
    writeln(outf,'Marginal totals read from file ',datfn,'.');
    for i:=1 to f do read(datf,rt[i]);
    readln(datf);
    if (not (graph or Eof(datf))) then begin
      for i:=1 to g do read(datf,ct[i]);
     end else begin
      g := f;
      for i:=1 to g do ct[i] := rt[i];
     end;
    if (f=g) then begin
      SameDegrees := true;
      for i:=1 to f do if (ct[i] <> rt[i]) then SameDegrees := false;
     end else SameDegrees := false;
   end else begin (* adjacency matrix file *)
    datmatrix := true;
    writeln(outf,
      'Marginal totals calculated from adjacency matrix in file ',datfn,'.');
    for i:=1 to fmax do rt[i] := 0;
    for j:=1 to gmax do ct[j] := 0;
    symm := true;
    with dmat do begin
      AllocateMat(f,g,ert);
      for i:=1 to f do begin
        for j:=1 to g do begin
          read(datf,dumij);
          if ((graph or digr) and (i=j)) then dumij := 0;
          Def(i,j,(dumij > 0));
          if (dumij > 0) then
            begin inc(rt[i]); inc(ct[j]);
             if ((j < i) and (not arc(j,i))) then symm := false;
           end;
         end; (* for j *)
        readln(datf);
        if (Eof(datf) and (i < f)) then begin
          writeln(outf,'Error: end of file ',datfn,'after reading line ',i);
          writeln(outf,'whereas ',f,' lines are expected.');
          errtext := 'Data file has not enough rows.';
          f := 0; g := 0;
          exit;
         end;
       end; (* for i *)
      DeAllocateMat;
     end; (* with dmat *)
    if (f=g) then begin
      write(outf,'Data matrix is');
      if (not symm) then write(outf,' not');
      writeln(outf,' symmetric.');
      SameDegrees := true;
      for i:=1 to f do if (rt[i] <> ct[i]) then SameDegrees := false;
     end; (* if f=g *)
   end; (* adjacency matrix file *)
  closefile(datf);

  if eof(inf) then
    writeln(outf,'Warning: unexpected end of input file.');
  readln(inf,dum);
  conn := ((dum > 0) and (graph or digr));
  mutfix := false;
  if digr then begin
    readln(inf,dum);
    mutfix := (dum > 0);
    if mutfix then readln(inf,presmuts);
   end;

  writeln(outf,'Requested mode:');
  if full then writeln(outf,'no structural zeros.');
  if digr then begin
     write(outf,'square matrix with structurally zero diagonal');
     writeln(outf,' (directed graph).'); end;
  if mutfix then writeln(outf,'Number of mutual dyads to be fixed at ', presmuts,'.');
  if graph then begin
     write(outf,'symmetric square matrix with structurally zero diagonal');
     writeln(outf,' (undirected graph).'); end;
  if gnrl then
     writeln(outf,'general pattern of structural zeros.');
  if conn then begin
     write(outf,'Enumeration is restricted to');
     if digr then write(outf,' weakly');
     write(outf,' connected');
     if digr then writeln(outf,' digraphs.') else writeln(outf,' graphs.');
    end;
  if MolloyReed then writeln(outf,'Molloy-Reed algorithm is used.');

(* Structural zeros *)
  with strzer do begin
    for i:=1 to fmax do for j:=1 to gmax do Def(i,j,false);
    if (graph or digr) then for i:=1 to fmax do Def(i,i,true);
   end; (* with*)

(* Input of structural zeros for general case *)
  CheckNumber := -99;
  if gnrl then begin
    readln(inf,datfn);
    CheckNumber := StrToIntDef(datfn,-99);
   end;
  if (gnrl and (CheckNumber = -99)) then begin
    (* datfn must be a file name *)
    writeln(outf,'Structural zeros read from file ',datfn,'.');
    if (not FileExists(datfn)) then begin
      writeln(outf,'Error: data file ',datfn,' is not found.');
      f:=0; g := 0;
      Errtext := 'File '+datfn+' is not found.';
      exit;
     end;
(* read data file: *)
    assign(datf,datfn);
    reset(datf);
    for i:=1 to fmax do rzer[i] := 0;
    for j:=1 to gmax do czer[j] := 0;
    iend := 0; jend := 0;
    with strzer do begin
      for i:= 1 to f do begin
        for j:= 1 to g do begin
          if (eof(datf) and (iend+jend = 0)) then begin
            iend := i; jend := j;
           end;
          read(datf,dum);
          Def(i,j,(dum >= 1));
          if Arc(i,j) then begin
            rzer[i] := rzer[i] + 1;
            czer[j] := czer[j] + 1;
           end; (* if *)
         end; (* j *)
        readln(datf);
       end; (* i *)
      closefile(datf);
      if (iend + jend > 0) then begin
        writeln(outf,'Unexpected end of file ',datfn,' when trying to read');
        writeln(outf,'position ',jend,' of line ',iend,'.');
       end;
   end; (* if CheckNumber = -99 *)
  if (gnrl and (CheckNumber <> -99)) then begin
    writeln(outf,'No file name given for structural zeros.');
    writeln(outf,'Consequently, no structural zeros are assumed.');
   end;
  if gnrl then begin
      writeln(outf,'Structural zeros in following pattern : ');
      writeln(outf);
      for i:= 1 to f do begin
        for j:=1 to g do if strzer.Arc(i,j) then
          write(outf,'1') else write(outf,'0');
        writeln(outf);
       end;
       writeln(outf);
      end; (* with strzer *)
   end; (* if gnrl *)

  if ((CheckNumber = -99) and eof(inf)) then
    writeln(outf,'Warning: unexpected end of input file.');
  if (CheckNumber = -99) then readln(inf,dum) else dum := CheckNumber;
  if ((dum >= 1) and (dum <= numversion(f,g)))
      then fversion := dum else fversion := 1;
  if (dum <> fversion) then
    writeln(outf,'The number ',dum,' is not allowed as the version of evaluated statistics.');
  writeln(outf,'Selected version of evaluated statistics is ',fversion,'.');

(* trivial checks *)
    if (graph and (not SameDegrees)) then begin
      graph := false; digr := true;
      writeln(outf,'Mode changed to digraph instead of graph.');
      mutfix := false;
     end;
    writeln(outf);
    writeln(outf,'vertices');
    for i:= 1 to f do write(outf,i:3,' ');
    if (g > f) then for i:=f+1 to g do write(outf,i:3,' ');
    writeln(outf);
    writeln(outf,'row totals');
    for i:= 1 to f do write(outf,rt[i]:3,' '); writeln(outf);
    writeln(outf,'column totals');
    for i:= 1 to g do write(outf,ct[i]:3,' '); writeln(outf);
    good := true;
    for i:=1 to f do
      if not ((rt[i] >= 0) and (rt[i] + rzer[i] <= g)) then
        begin
        write(outf,'row total ', i, ' = ', rt[i], '.');
        writeln(outf,' This is impossible !'); good := false;
        errtext := 'Row total has impossible value.';
       end;
    for i:=1 to g do
      if not ((ct[i] >= 0) and (ct[i] + czer[i] <= f)) then
        begin
        write(outf,'column total ', i, ' = ', ct[i], '.');
        writeln(outf,' This is impossible !'); good := false;
        errtext := 'Column total has impossible value.';
       end;
   if  (sumc(rt,f) <> sumr(ct,g)) then
     begin
       writeln(outf,'The row and column totals do not have the same sum.');
       writeln(outf,'This is impossible.');
       errtext := 'Row and column totals incompatible.';
       good := false;
     end;
   if ((graph) and odd(sumc(rt,g))) then begin
     writeln(outf,
       'The degree sum is odd, which is impossible for an undirected graph.');
     errtext := 'No graph with these degrees.';  
     good := false;
    end;
   if (mutfix) then
     if ((presmuts < 0) or (2*presmuts > sumc(rt,f))) then begin
       writeln(outf,'The precribed number of mutual dyads is ',presmuts,'.');
       writeln(outf,'This is impossible.');
       errtext := 'Prescribed number of mutual dyads impossible.';
       good := false;
     end;
   if (not good) then begin
     f := 0; g := 0;
     exit;
    end;

(* at this point, the readinfile procedure in ZOELIB is terminated. *)
  rorder := (graph or digr);

  readln(inf,dum);
  if (dum >= 0) then nrun := dum else nrun := 100;
  writeln(outf,dum,' simulations requested.');
  if (nrun = 1) then nrun := 2;
  if (dum <> nrun) then writeln(outf,' This is changed to ',nrun,'.');
  if eof(inf) then
    writeln(outf,'Warning: unexpected end of input file.');

(* linear combinations of triad counts *)
  if (graph or digr) then begin
    readln(inf,dum);
    if (dum >= 0) and (dum <= numlmax) then numl := dum else numl := 0;
    readln(inf,dum);
    if (dum in [1..2]) then option := dum else option := 1;
    if (numl = 0) then option := 1;
    if (numl >= 1) then begin
      if (option <=1) then
        for i:=1 to numl do begin
          for j:=1 to 16 do if (i = j)
            then weit[i,j]:= 1.0 else weit[i,j] := 0.0;
          weit[i,0] := 0.1;
 (* These weit[i,0] are arbitrary values with which the
    generated linear combinations will be compared.
    They may be replaced by anything. *)
         end
       else (* option =  2 *) begin
        readln(inf,datfn);
        writeln(outf,
  'Weights for linear combinations of triad counts read from file ',datfn,'.');
        if (not FileExists(datfn)) then begin
          writeln(outf,'Error: data file ',datfn,' is not found.');
          f:=0; g := 0;
          Errtext := 'File '+datfn+' not found.';
          exit;
         end;
        assign(datf,datfn);
        reset(datf);
        anycrit := false;
        writeln(outf,'Weights read are the following : ');
        for i:= 1 to numl do begin
          for j:= 1 to 16 do begin
             read(datf,weit[i,j]);
             write(outf,' ',weit[i,j]:5:1);
             end; (* j *)
          if eoln(datf) then begin
            weit[i,0] := 1;
            crit[i] := false;
            readln(datf);
           end else begin
            readln(datf,ts);
            if parsestring(ts,x) then begin
              weit[i,0] := x;
              crit[i] := true;
              anycrit := true;
             end else begin
              weit[i,0] := 1;
              crit[i] := false;
             end;
           end;
          writeln(outf);
         end; (* i *)
        close(datf);
        if anycrit then begin
          writeln(outf,'with critical values :');
          for i:=1 to numl do
            if crit[i] then writeln(outf,i:6,' ',weit[i,0]:7:4);
          writeln(outf);
         end;
       end; (* if option >= 2 *)
     end; (* if numl >= 1 *)
   end (* if graph or digr *) else numl := 0;

 if printall then begin
   pname := proname+'.prn';
   assign(pfile,pname);
   if FileExists(pname) then begin
     append(pfile);
     writeln(pfile);
    end else rewrite(pfile);
   printfileconn := true;
   writeln(pfile,'This file contains all found 0-1 matrices ');
   writeln(pfile,'with the desired row and column totals.');
   writeln(pfile,'Mutual dyads are indicated by M.');
   writeln(pfile);
   write(pfile,'Row totals :    ');
   for i:=1 to f do
     begin write(pfile,rt[i]:3);
     if ((i mod 20) = 0) then writeln(pfile); end;
   writeln(pfile); write(pfile,'Column totals : ');
   for i:=1 to g do
     begin write(pfile,ct[i]:3);
     if ((i mod 20) = 0) then writeln(pfile); end;
   writeln(pfile);
   if graph then writeln(pfile,'Undirected graph.');
   if digr then writeln(pfile,'Directed graph.');
   if full then writeln(pfile,'No structural zeros.');
   if mutfix then writeln(pfile,'Number of mutuals fixed at ', presmuts);
   if gnrl then begin
     writeln(pfile,'Structural zeros in following pattern : ');
     writeln(pfile);
     for i:= 1 to f do begin
       for j:=1 to g do if strzer.arc(i,j) then
         write(pfile,'0') else write(pfile,'1');
       writeln(pfile);
      end;
     writeln(pfile);
    end;
  end; (* if printall *)

  if full then mode := 1;
  if digr then mode := 2;
  if graph then mode := 3;
  if gnrl then mode := 4;
  if (digr and mutfix) then mode := 5;
  if rorder then mode := mode+10;
  if MolloyReed then mode := mode + 100;

  closefile(inf);

  if (datmatrix) then begin
    Func(dmat,f,g,numl,fversion,graph,digr,conn,0,rt,weit,f1,f2,rf1);
    heading(3,outf,'Function values observed for this data matrix');
    for i:=1 to NumFunc(fversion,f,g) do begin
      ts := Tex(i,0,fversion,graph,digr,conn,true);
      if length(ts) < 60 then for j:=length(ts) to 59 do ts := ts+' ';
      writeln(outf,ts,f1[i]:10);
     end; (* i *)
    writeln(outf);
    if (numl > 0) then begin
      writeln(outf,'Observed triad census ');
      triadcount(dmat,f,tc);
      for i:=1 to 16 do writeln(outf,i:2,'. ',triadname(triadkind[i]),' ',
                                tc[i]:7);
      writeln(outf);
      writeln(outf,'Observed linear combinations of triad counts');
      for i:=1 to numl do begin
        writeln(outf,i:4,'. ',f2[i]:10:4);
        weit[i,0] := f2[i];
       end;
      writeln(outf);
     end; (* if numl > 0 *)
   end; (* if datmatrix *)
end; (* readinfile *)


procedure Extremes(var mean: Res; var m3,m4: ResI);
var h: Byte; i: Integer;
begin
  for h:=1 to numf do begin
    m3[h] := m2; m4[h] := 0;
    for i:=0 to m2 do
      if (mean[h,i] > 1e-30) then begin
        if (m3[h] > i) then m3[h] := i;
        if (m4[h] < i) then m4[h] := i;
       end; (* if *)
   end; (* h *)
end; (* Extremes *)


procedure Report1(var fl: Text; m3,m4:Integer; var mean,se: Res1; zer: integer);
(* mean and se are estimates and their standard errors *)
var i,row,nrows,rowmin,rowleng: Integer;
begin
   if (m3 > m4) then begin
     writeln(fl,'Minimum index is ',m3,' while maximum is ',m4,' .');
     writeln(fl,'No report.');
     Exit;
    end; (* if *)
   nrows := ((m4-m3) div 10) + 1;
   for row := 1 to nrows do begin
     rowmin := m3 + (row-1)*10 - 1;
     rowleng := m4 - m3 - row*10 + 11;
     if (rowleng > 10) then rowleng := 10;
     write(fl,'  ',rowmin+1+zer:3);
     for i:= 2 to rowleng do write(fl,rowmin + i + zer:7); writeln(fl);
     for i:= 1 to rowleng do write(fl,' ',mean[rowmin+i]:6:4); writeln(fl);
     for i:= 1 to rowleng do write(fl,' ',se[rowmin+i]:6:4); writeln(fl);
     writeln(fl);
    end; (* rows *)
  writeln(fl);
  end (* Report1 *);

procedure Report2(var fl: Text; var mean,sd: ResL; var cov: ResL2);
(* Report for linear combinations of triad counts. *)
var h,hh,l1,line: Byte;
begin
  if (numl <= 0) then Exit;
  Heading(3,fl,'Linear combinations of triad counts');
  writeln(fl);
  writeln(fl,'Weights'); writeln(fl);
  write(fl,' #   ');
  for hh:=1 to numl do
    begin write(fl,hh:7); if (hh = 8) then
      begin writeln(fl); write(fl,'      '); end; end;
  writeln(fl);
  for h:=1 to 16 do begin
    write(fl,Triadname(Triadkind[h]),'   ');
    for hh :=1 to numl do
      begin write(fl,' ',weit[hh,h]:6:3); if (hh = 8) then
        begin writeln(fl); write(fl,'       '); end; end;
      writeln(fl);
   end;
  writeln(fl);
  writeln(fl,'Critical values'); write(fl,'         ');
  for hh:=1 to numl do
    begin write(fl,' ',weit[hh,0]:6:4); if (hh = 8) then
      begin writeln(fl); write(fl,'       '); end;
     end;
  writeln(fl);
  writeln(fl,'Estimated means (with standard errors) of linear combinations.');
  line := 0;
  write(fl,'     ');
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,8*line+h:8); writeln(fl);
    inc(line);
    write(fl,'     ');
   until (8*line >= numl);
  writeln(fl);
  write(fl,'mean    ');
  line := 0;
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,' ',mean[8*line + h]:7:4); writeln(fl);
    inc(line);
    write(fl,'        ');
   until (8*line >= numl);
  writeln(fl);
  write(fl,'s.e.    ');
  line := 0;
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,' ',sd[8*line + h]:7:4); writeln(fl);
    inc(line);
    write(fl,'        ');
   until (8*line >= numl);
  writeln(fl);
  writeln(fl,'Covariance matrix of linear combinations');
  write(fl,'  ');
  line := 0;
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,8*line+h:10); writeln(fl);
    inc(line);
    write(fl,'       ');
   until (8*line >= numl);
  writeln(fl);
  for h:=1 to numl do begin
    write(fl,h:3,'    ');
    line := 0;
    repeat
      l1 := numl - 8*line;
      if (l1 > 8) then l1 := 8;
      for hh:=1 to l1 do write(fl,' ',cov[h,8*line + hh]:9:6); writeln(fl);
      inc(line);
      write(fl,'       ');
     until (8*line >= numl);
    writeln(fl);
   end; (* h *)
  writeln(fl);
  writeln(fl,'Exceedance probabilities (p-values)');
  writeln(fl,'(i.e., probabilities that linear combinations are at least as large');
  writeln(fl,'       as the critical values)');
  line := 0;
  write(fl,'              ');
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,8*line+h:7); writeln(fl);
    inc(line);
    write(fl,'             ');
   until (8*line >= numl);
  writeln(fl);
  write(fl,'p-values        ');
  line := 0;
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,' ',mean[numl + 8*line + h]:6:3); writeln(fl);
    inc(line);
    write(fl,'                ');
   until (8*line >= numl);
  writeln(fl);
  write(fl,'standard errors ');
  line := 0;
  repeat
    l1 := numl - 8*line;
    if (l1 > 8) then l1 := 8;
    for h:=1 to l1 do write(fl,' ',sd[numl + 8*line + h]:6:3); writeln(fl);
    inc(line);
    write(fl,'                ');
   until (8*line >= numl);
  writeln(fl);
end; (* Report2 *)


procedure Giveresult(var outf: textfile);
var rang,barnum,ii,j: Integer;
    h, i: Byte;
    m3,m4: ResI;
    mean1, sd1: Res1;
    summean,zero: Ress;
    hzero: integer;
    full,digr,graph,gnrl,mutfix,rorder,connect,MolReed: Boolean;
begin
  if (nsuc <= 0) then exit;
(* Results for the probability distribution of the numf functions *)
  Extremes(mean,m3,m4);
  unpackmode(full,digr,graph,gnrl,mutfix,rorder,connect,MolReed);
  for h:=1 to numf do begin
    if (h=1) then hzero := mutzero else hzero := 0;
    heading(3,outf,Tex(h,0,fversion,graph,digr,connect,true));
    write(outf,'Estimated mean               ', hzero + mean[h,-1]:8:4);
    writeln(outf,'     (standard error ', sd[h,-1]:8:4, ' )');
    writeln(outf,'Estimated variance           ', mean[h,-2]:8:4);
    writeln(outf,'Estimated standard deviation ', root(mean[h,-2]):8:4);
    if small[h] then begin
      writeln(outf,'At least one function value was smaller than ',hzero,'.');
      writeln(outf,'All smaller values were set to ',hzero,'.');
     end; (* if *)
    if large[h] then begin
      writeln(outf,'At least one function value was larger than ',m2+hzero,' .');
     end; (* if *)
    if (m3[h] = m4[h]) then
      writeln(outf,'All observed values equal to ',m3[h]+hzero,'.');
    if (m3[h] < m4[h]) then begin
      rang := ((m4[h]-m3[h]) div 70) + 1;
      if (rang <= 0) then rang := 1;
      barnum := (m4[h]-m3[h]+1) div rang;
      if rang*barnum < m4[h]-m3[h]+1 then barnum := barnum + 1;
      j := m3[h]-1;
      for ii:=1 to barnum do begin
        summean[ii] := 0.0;
        zero[ii] := 0.0;
        for i := 1 to rang do begin
          j := j+1;
          if (j <= m2) then summean[ii] := summean[ii]+mean[h,j];
         end; (* i *)
        end;(* ii *)
      if (m4[h] + hzero < 100) then begin
        Barchart(outf,1,rang,barnum,m3[h]+hzero,summean,zero);
       end else begin
        if (hzero <> 0) then begin
          writeln(outf);
          writeln(outf,'Barchart for ');
          writeln(outf,Tex(h,hzero,fversion,graph,digr,connect,true));
         end;
        Barchart(outf,1,rang,barnum,m3[h],summean,zero);
       end;
      writeln(outf);
      for ii :=0 to m2 do
        begin mean1[ii] := cummean[h,ii]; sd1[ii] := cumsd[h,ii]; end;
      writeln(outf,'Minimum value found ',m3[h]+hzero,'; maximum ',m4[h]+hzero,' .');
      writeln(outf);
      writeln
       (outf,'Estimated cumulative ( <= ) probabilities with standard errors.');
      Report1(outf,m3[h],m4[h],mean1,sd1,hzero);
     end; (* if m3 < m4 *)
   end; (* h *)
  Report2(outf,meanl,sdl,covl);
end;      (* Giveresult *)


procedure Result0(var ff: Text; mode: byte);
var i: Byte;
begin  
  if (nrun = 0) then begin
    writeln(ff,'No simulations.');
   end else begin
    if ChangedOrder then begin
      writeln(ff,'For efficiency purposes, rows and columns are similarly reordered.');
      writeln(ff,'Look below at the row and column sums to see the new order.');
     end;
    writeln(ff);
    Heading(2,ff,'Simulation results');
    if (nrun >= 2) then
      writeln(ff,'Number of runs ', nrun, ';  number of trials ', ntry, ' .');
    writeln(ff);
    writeln(ff,'Vertices : ');
    for i:=1 to f do begin
      write(ff,i:3);
      if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
    if (g > f) then for i:=f+1 to g do begin
      write(ff,i:3);
      if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
    writeln(ff);
    writeln(ff,'Row totals : ');
    for i:=1 to f do begin
      write(ff,rt[i]:3);
      if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
    writeln(ff);
    writeln(ff,'Column totals : ');
    for i:=1 to g do begin
      write(ff,ct[i]:3);
      if (i div 20) >= (i/20.0 - 0.001) then writeln(ff); end;
    writeln(ff);
    if (nrun >= 2) and (not UseResfile) then begin
     if (root(varp*nrun) > 2.0) then begin
      write(ff,'The standard deviation of 1/probabilities is ');
      writeln(ff,root(varp*nrun):10:2);
      writeln(ff,'If this standard deviation is very high, then the simulation procedure');
      writeln(ff,'can be unreliable.');
      writeln(ff,'(50 is high; 100 is very high.)');
      writeln(ff);
      end;
     end;
   end; (* if nrun > 0 *)
end; (* Result0 *)


begin (* unit ZoSimlib *)
end.
